/*
 * Copyright (c) 2012-2013, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *  ======== AppDsp.c ========
 *  Platform: OMAPL138_linux_elf
 */

/* this define must precede inclusion of any xdc header file */
#define Registry_CURDESC AppDsp__Desc
#define MODULE_NAME "AppDsp"

#define HEAPID 0

#include <xdc/std.h>
#include <xdc/runtime/Diags.h>
#include <xdc/runtime/Registry.h>
#include <xdc/runtime/System.h>

#include <ti/ipc/MessageQ.h>

#include <ti/sysbios/knl/Semaphore.h>
#include <ti/sysbios/knl/Task.h>

/* local header files */
#include "AppDsp.h"
#include "../shared/SvrMsg.h"

/* module structure */
typedef struct {
    MessageQ_Handle     localMsgQue;    /* created locally */
    MessageQ_QueueId    svrQue[10];     /* opened remotely */
    SvrMsg_Msg *        msg;
} AppDsp_Module;

/* friend data */
extern Semaphore_Handle SvrDsp_doneFlag;

/* private data */
Registry_Desc           Registry_CURDESC;
static AppDsp_Module   Module;


/*
 *  ======== AppDsp_destroy ========
 */
Int AppDsp_destroy(Void)
{
    Int     status = 0;
    UInt16  numProcs;
    UInt16  procId;

    System_printf("--> AppDsp_destroy:\n");

    /* TODO enter gate */

    /* wait for server thread to finish */
    while (Semaphore_getCount(SvrDsp_doneFlag) > 0) {
        Task_sleep(1);
    }


    numProcs = MultiProc_getNumProcessors();
    for (procId = 0; procId < numProcs; procId++) {

        /* fill and send message */
        Module.msg->cmd = SvrMsg_Cmd_STOP;
        MessageQ_put(Module.svrQue[procId], (MessageQ_Msg)Module.msg);

    }


    /* close each remote processor's server queue */
    numProcs = MultiProc_getNumProcessors();
    for (procId = 0; procId < numProcs; procId++) {

        /* close remote message queue */
        status = MessageQ_close(&Module.svrQue[procId]);

        if (status < 0) {
            System_printf("AppDsp_destroy: procId=%d, server queue close "
                    "error=%d \n", procId, status);
            goto leave;
        }
    }

    /* delete the local message queue */
    status = MessageQ_delete(&Module.localMsgQue);

    if (status < 0) {
        System_printf("AppDsp_destroy: queue delete error=%d \n", status);
        goto leave;
    }

leave:
    /* disable log events */
    System_printf("<-- AppDsp_destroy: status=%d \n", status);
    Diags_setMask(MODULE_NAME"-FEX");

    /*
     * Note that there isn't a Registry_removeModule() yet:
     *     https://bugs.eclipse.org/bugs/show_bug.cgi?id=315448
     *
     * ... but this is where we'd call it.
     */

    /* TODO leave gate */
    return(status);
}

/*
 *  ======== AppDsp_run ========
 */
Int AppDsp_run(Void)
{
    Int     status = 0;
    Int     count;
    UInt16  numProcs;
    UInt16  procId;

    System_printf("--> AppDsp_run: \n");

    /* ping each remote server */
    numProcs = MultiProc_getNumProcessors();
    for (procId = 0; procId < numProcs; procId++) {
        for (count = 1; count <= SvrMsg_NUM_MSGS; count++) {
            System_printf("AppDsp_run: ping procId=%d \n", procId);

            /* fill and send message */
            Module.msg->cmd = SvrMsg_Cmd_START;
            Module.msg->svrProcId = MultiProc_INVALIDID;
            MessageQ_put(Module.svrQue[procId], (MessageQ_Msg)Module.msg);

            /* wait for return message */
            status = MessageQ_get(Module.localMsgQue,
                    (MessageQ_Msg *)&Module.msg, MessageQ_FOREVER);

            if (status < 0) {
                System_printf("AppDsp_run: message receive error=%d \n",
                        status);
                goto leave;
            }
            else if (Module.msg->svrProcId != procId) {
                System_printf("AppDsp_run: message receive error=%d \n",
                        status);
                goto leave;
            }

            System_printf("AppDsp_run: ack received, procId=%d",
                    procId);
        }
    }

leave:
    System_printf("<-- AppDsp_run: %d \n", status);
    return(status);
}

/*
 *  ======== AppDsp_setup ========
 */
Int AppDsp_setup(Void)
{
    Int                 status = 0;
    Registry_Result     result;
    MessageQ_Params     msgqParams;
    UInt16              numProcs;
    UInt16              procId;
    Char                cbuf[48];

    /* TODO enter gate */


    /* initialize module state */
    Module.localMsgQue = NULL;
    Module.msg = NULL;

    /* register with xdc.runtime to get a diags mask */
    result = Registry_addModule(&Registry_CURDESC, MODULE_NAME);

    if ((result != Registry_SUCCESS) && (result != Registry_ALREADY_ADDED)) {
        status = -1;
        goto leave;
    }

    /* enable some trace */
    Diags_setMask(MODULE_NAME"+FEX");
    System_printf("--> AppDsp_setup:\n");

    /* create local message queue (inbound messages) */
    MessageQ_Params_init(&msgqParams);
    Module.localMsgQue = MessageQ_create(Dsp_AppMsgQue, &msgqParams);

    if (Module.localMsgQue == NULL) {
        System_printf("AppDsp_setup: message queue create failed");
        status = -1;
        goto leave;
    }

    /* open each remote processor's server queue */
    numProcs = MultiProc_getNumProcessors();
    for (procId = 0; procId < numProcs; procId++) {

        /* compute the remote processor server queue name */
        System_sprintf(cbuf, SvrMsg_SvrQueNameFmt, MultiProc_getName(procId));

        /* open the remote message queue, loop until queue is available */
        do {
            status = MessageQ_open(cbuf, &Module.svrQue[procId]);
            if (status == MessageQ_E_NOTFOUND) {
                Task_sleep(1);
            }
        } while (status == MessageQ_E_NOTFOUND);

        if (status < 0) {
            System_printf("AppDsp_setup: could not open slave message queue \n");
            goto leave;
        }

        System_printf("AppDsp_setup: procId=%d, opened server queue \n",
                procId);
    }

     /* allocate a static message */
    Module.msg = (SvrMsg_Msg *)MessageQ_alloc(HEAPID,sizeof(SvrMsg_Msg));

    if (Module.msg == NULL) {
        System_printf("AppDsp_setup: could not allocate message\n");
        status = -1;
        goto leave;
    }

    /* initialize message header */
    MessageQ_setReplyQueue(Module.localMsgQue, (MessageQ_Msg)Module.msg);

    System_printf("AppDsp_setup: slave is ready \n");

leave:
    /* TODO leave gate */
    System_printf("<-- AppDsp_setup:\n");
    return(status);
}
